// Copyright (C) 1999-2000 Id Software, Inc.
//
/*
=======================================================================

USER INTERFACE MAIN

=======================================================================
*/


/*
The work contained within this file is software written by various
copyright holders.  The initial contributor, Id Software holds all
copyright over their software.  However, software used and written by
and for UI Enhanced has copyrights held by the initial author of the
software.

The changes written by and for UI Enhanced are contained alongside the
original work from Id Software for convenience and ease of interoperability.

For the code contained herein that was written by Id Software, see
the license agreement on their original archive for restrictions and
limitations.

The UI Enhanced copyright owner permit free reuse of his code contained
herein, as long as the following terms are met:

----------------------------------------------------------------------
1) Credit is given in a place where users of the mod may read
   it. (Title screen, credit screen or README will do).  The
   recommended format is: "First, Last, alias, email"

2) There are no attempts to misrepresent the public as to who made the
   alterations.  The UI Enhanced copyright owner does not give permission
   for others to release software under the UI Enhanced name.
----------------------------------------------------------------------

Ian Jefferies   - HypoThermia ( uie@planetquake.com )
http://www.planetquake.com/uie
*/

#include "ui_local.h"


/*
================
vmMain

This is the only way control passes into the module.
This must be the very first function compiled into the .qvm file
================
*/
int vmMain( int command, int arg0, int arg1, int arg2, int arg3, int arg4, int arg5, int arg6, int arg7, int arg8, int arg9, int arg10, int arg11  ) {
	switch ( command ) {
	case UI_GETAPIVERSION:
		return UI_API_VERSION;

	case UI_INIT:
		UI_Init();
		return 0;

	case UI_SHUTDOWN:
		UI_Shutdown();
		return 0;

	case UI_KEY_EVENT:
		UI_KeyEvent( arg0, arg1 );
		return 0;

	case UI_MOUSE_EVENT:
		UI_MouseEvent( arg0, arg1 );
		return 0;

	case UI_REFRESH:
		UI_Refresh( arg0 );
		return 0;

	case UI_IS_FULLSCREEN:
		return UI_IsFullscreen();

	case UI_SET_ACTIVE_MENU:
		UI_SetActiveMenu( arg0 );
		return 0;

	case UI_CONSOLE_COMMAND:
		return UI_ConsoleCommand(arg0);

	case UI_DRAW_CONNECT_SCREEN:
		UI_DrawConnectScreen( arg0 );
		return 0;
	case UI_HASUNIQUECDKEY:				// mod authors need to observe this
		return qfalse;
	}

	return -1;
}


/*
================
cvars
================
*/

typedef struct {
	vmCvar_t	*vmCvar;
	char		*cvarName;
	char		*defaultString;
	int			cvarFlags;
} cvarTable_t;

vmCvar_t	ui_ffa_fraglimit;
vmCvar_t	ui_ffa_timelimit;

vmCvar_t	ui_tourney_fraglimit;
vmCvar_t	ui_tourney_timelimit;

vmCvar_t	ui_team_fraglimit;
vmCvar_t	ui_team_timelimit;
vmCvar_t	ui_team_friendly;

vmCvar_t	ui_ctf_capturelimit;
vmCvar_t	ui_ctf_timelimit;
vmCvar_t	ui_ctf_friendly;

vmCvar_t	ui_arenasFile;
vmCvar_t	ui_botsFile;
vmCvar_t	ui_spScores1;
vmCvar_t	ui_spScores2;
vmCvar_t	ui_spScores3;
vmCvar_t	ui_spScores4;
vmCvar_t	ui_spScores5;
vmCvar_t	ui_spAwards;
vmCvar_t	ui_spVideos;
vmCvar_t	ui_spSkill;

vmCvar_t	ui_spSelection;

vmCvar_t	ui_browserMaster;
vmCvar_t	ui_browserGameType;
vmCvar_t	ui_browserSortKey;
vmCvar_t	ui_browserShowFull;
vmCvar_t	ui_browserShowEmpty;

vmCvar_t	ui_brassTime;
vmCvar_t	ui_drawCrosshair;
vmCvar_t	ui_drawCrosshairNames;
vmCvar_t	ui_marks;

vmCvar_t	ui_server1;
vmCvar_t	ui_server2;
vmCvar_t	ui_server3;
vmCvar_t	ui_server4;
vmCvar_t	ui_server5;
vmCvar_t	ui_server6;
vmCvar_t	ui_server7;
vmCvar_t	ui_server8;
vmCvar_t	ui_server9;
vmCvar_t	ui_server10;
vmCvar_t	ui_server11;
vmCvar_t	ui_server12;
vmCvar_t	ui_server13;
vmCvar_t	ui_server14;
vmCvar_t	ui_server15;
vmCvar_t	ui_server16;

vmCvar_t	ui_cdkeychecked;

// HypoThermia: UI Enhanced cvars
vmCvar_t	uie_gametype;	// g_gametype is latched in a different vm
vmCvar_t	uie_pure;
vmCvar_t	uie_dedicated;
vmCvar_t	uie_inactivity;
vmCvar_t	uie_animsfx;
vmCvar_t	uie_mapicons;
vmCvar_t	uie_autoclosebotmenu;
vmCvar_t	uie_config_showid;
vmCvar_t	uie_pmove_fixed;
vmCvar_t	uie_pmove_msec;
vmCvar_t	uie_smoothclients;
vmCvar_t	uie_syncclients;
vmCvar_t	uie_allowmaxrate;
vmCvar_t	uie_maxrate;
vmCvar_t	uie_allowdownload;
vmCvar_t	uie_allowpass;
vmCvar_t	uie_password;
vmCvar_t	uie_allowvote;
vmCvar_t	uie_minPing;
vmCvar_t	uie_maxPing;
vmCvar_t	uie_allowMinPing;
vmCvar_t	uie_allowMaxPing;
vmCvar_t	uie_preventConfigBug;
vmCvar_t	uie_map_multisel;
vmCvar_t	uie_map_list;
vmCvar_t	uie_bot_multisel;
vmCvar_t	uie_bot_list;

vmCvar_t	uie_ingame_dynamicmenu;

vmCvar_t	uie_ffa_fragtype;
vmCvar_t	uie_ffa_timetype;
vmCvar_t	uie_ffa_customfraglimits;
vmCvar_t	uie_ffa_customtimelimits;
vmCvar_t	uie_ffa_maplist;
vmCvar_t	uie_ffa_maplistexclude;
vmCvar_t	uie_ffa_MapRepeat;
vmCvar_t	uie_ffa_MapSource;
vmCvar_t	uie_ffa_RandomMapCount;
vmCvar_t	uie_ffa_RandomMapType;
vmCvar_t	uie_ffa_slottype;
vmCvar_t	uie_ffa_botname;
vmCvar_t	uie_ffa_botexclude;
vmCvar_t	uie_ffa_botskill;
vmCvar_t	uie_ffa_BotSelection;
vmCvar_t	uie_ffa_BotCount;
vmCvar_t	uie_ffa_BotChange;
vmCvar_t	uie_ffa_OpenSlotCount;
vmCvar_t	uie_ffa_BotSkillType;
vmCvar_t	uie_ffa_BotSkillValue;
vmCvar_t	uie_ffa_BotSkillBias;
vmCvar_t	uie_ffa_PlayerJoinAs;
vmCvar_t	uie_ffa_hostname;
vmCvar_t	uie_ffa_ForceRespawn;
vmCvar_t	uie_ffa_itemGroups;
vmCvar_t	uie_ffa_itemsHidden;
vmCvar_t	uie_ffa_doWarmUp;
vmCvar_t	uie_ffa_WarmUp;

vmCvar_t	uie_tourney_fragtype;
vmCvar_t	uie_tourney_timetype;
vmCvar_t	uie_tourney_customfraglimits;
vmCvar_t	uie_tourney_customtimelimits;
vmCvar_t	uie_tourney_maplist;
vmCvar_t	uie_tourney_maplistexclude;
vmCvar_t	uie_tourney_MapRepeat;
vmCvar_t	uie_tourney_MapSource;
vmCvar_t	uie_tourney_RandomMapCount;
vmCvar_t	uie_tourney_RandomMapType;
vmCvar_t	uie_tourney_slottype;
vmCvar_t	uie_tourney_botname;
vmCvar_t	uie_tourney_botexclude;
vmCvar_t	uie_tourney_botskill;
vmCvar_t	uie_tourney_BotSelection;
vmCvar_t	uie_tourney_BotCount;
vmCvar_t	uie_tourney_BotChange;
vmCvar_t	uie_tourney_OpenSlotCount;
vmCvar_t	uie_tourney_BotSkillType;
vmCvar_t	uie_tourney_BotSkillValue;
vmCvar_t	uie_tourney_BotSkillBias;
vmCvar_t	uie_tourney_PlayerJoinAs;
vmCvar_t	uie_tourney_hostname;
vmCvar_t	uie_tourney_ForceRespawn;
vmCvar_t	uie_tourney_itemGroups;
vmCvar_t	uie_tourney_itemsHidden;
vmCvar_t	uie_tourney_doWarmUp;
vmCvar_t	uie_tourney_WarmUp;

vmCvar_t	uie_team_fragtype;
vmCvar_t	uie_team_timetype;
vmCvar_t	uie_team_customfraglimits;
vmCvar_t	uie_team_customtimelimits;
vmCvar_t	uie_team_maplist;
vmCvar_t	uie_team_maplistexclude;
vmCvar_t	uie_team_MapRepeat;
vmCvar_t	uie_team_MapSource;
vmCvar_t	uie_team_RandomMapCount;
vmCvar_t	uie_team_RandomMapType;
vmCvar_t	uie_team_slottype;
vmCvar_t	uie_team_botname;
vmCvar_t	uie_team_botexclude;
vmCvar_t	uie_team_botskill;
vmCvar_t	uie_team_BotSelection;
vmCvar_t	uie_team_BotCount;
vmCvar_t	uie_team_BotChange;
vmCvar_t	uie_team_OpenSlotCount;
vmCvar_t	uie_team_BotSkillType;
vmCvar_t	uie_team_BotSkillValue;
vmCvar_t	uie_team_BotSkillBias;
vmCvar_t	uie_team_PlayerJoinAs;
vmCvar_t	uie_team_TeamSwapped;
vmCvar_t	uie_team_hostname;
vmCvar_t	uie_team_ForceRespawn;
vmCvar_t	uie_team_AutoJoin;
vmCvar_t	uie_team_TeamBalance;
vmCvar_t	uie_team_itemGroups;
vmCvar_t	uie_team_itemsHidden;
vmCvar_t	uie_team_doWarmUp;
vmCvar_t	uie_team_WarmUp;

vmCvar_t	uie_ctf_capturetype;
vmCvar_t	uie_ctf_timetype;
vmCvar_t	uie_ctf_customcapturelimits;
vmCvar_t	uie_ctf_customtimelimits;
vmCvar_t	uie_ctf_maplist;
vmCvar_t	uie_ctf_maplistexclude;
vmCvar_t	uie_ctf_MapRepeat;
vmCvar_t	uie_ctf_MapSource;
vmCvar_t	uie_ctf_RandomMapCount;
vmCvar_t	uie_ctf_RandomMapType;
vmCvar_t	uie_ctf_slottype;
vmCvar_t	uie_ctf_botname;
vmCvar_t	uie_ctf_botexclude;
vmCvar_t	uie_ctf_botskill;
vmCvar_t	uie_ctf_BotSelection;
vmCvar_t	uie_ctf_BotCount;
vmCvar_t	uie_ctf_BotChange;
vmCvar_t	uie_ctf_OpenSlotCount;
vmCvar_t	uie_ctf_BotSkillType;
vmCvar_t	uie_ctf_BotSkillValue;
vmCvar_t	uie_ctf_BotSkillBias;
vmCvar_t	uie_ctf_PlayerJoinAs;
vmCvar_t	uie_ctf_TeamSwapped;
vmCvar_t	uie_ctf_hostname;
vmCvar_t	uie_ctf_ForceRespawn;
vmCvar_t	uie_ctf_AutoJoin;
vmCvar_t	uie_ctf_TeamBalance;
vmCvar_t	uie_ctf_itemGroups;
vmCvar_t	uie_ctf_itemsHidden;
vmCvar_t	uie_ctf_doWarmUp;
vmCvar_t	uie_ctf_WarmUp;
// HypoThermia: end Enhanced UI cvars


cvarTable_t		cvarTable[] = {
	{ &ui_ffa_fraglimit, "ui_ffa_fraglimit", "20", CVAR_ARCHIVE },
	{ &ui_ffa_timelimit, "ui_ffa_timelimit", "0", CVAR_ARCHIVE },

	{ &ui_tourney_fraglimit, "ui_tourney_fraglimit", "0", CVAR_ARCHIVE },
	{ &ui_tourney_timelimit, "ui_tourney_timelimit", "15", CVAR_ARCHIVE },

	{ &ui_team_fraglimit, "ui_team_fraglimit", "0", CVAR_ARCHIVE },
	{ &ui_team_timelimit, "ui_team_timelimit", "20", CVAR_ARCHIVE },
	{ &ui_team_friendly, "ui_team_friendly",  "1", CVAR_ARCHIVE },

	{ &ui_ctf_capturelimit, "ui_ctf_capturelimit", "8", CVAR_ARCHIVE },
	{ &ui_ctf_timelimit, "ui_ctf_timelimit", "30", CVAR_ARCHIVE },
	{ &ui_ctf_friendly, "ui_ctf_friendly",  "0", CVAR_ARCHIVE },

	{ &ui_arenasFile, "g_arenasFile", "", CVAR_INIT|CVAR_ROM },
	{ &ui_botsFile, "g_botsFile", "", CVAR_INIT|CVAR_ROM },
	{ &ui_spScores1, "g_spScores1", "", CVAR_ARCHIVE | CVAR_ROM },
	{ &ui_spScores2, "g_spScores2", "", CVAR_ARCHIVE | CVAR_ROM },
	{ &ui_spScores3, "g_spScores3", "", CVAR_ARCHIVE | CVAR_ROM },
	{ &ui_spScores4, "g_spScores4", "", CVAR_ARCHIVE | CVAR_ROM },
	{ &ui_spScores5, "g_spScores5", "", CVAR_ARCHIVE | CVAR_ROM },
	{ &ui_spAwards, "g_spAwards", "", CVAR_ARCHIVE | CVAR_ROM },
	{ &ui_spVideos, "g_spVideos", "", CVAR_ARCHIVE | CVAR_ROM },
	{ &ui_spSkill, "g_spSkill", "2", CVAR_ARCHIVE | CVAR_LATCH },

	{ &ui_spSelection, "ui_spSelection", "", CVAR_ROM },

	{ &ui_browserMaster, "ui_browserMaster", "0", CVAR_ARCHIVE },
	{ &ui_browserGameType, "ui_browserGameType", "0", CVAR_ARCHIVE },
	{ &ui_browserSortKey, "ui_browserSortKey", "4", CVAR_ARCHIVE },
	{ &ui_browserShowFull, "ui_browserShowFull", "1", CVAR_ARCHIVE },
	{ &ui_browserShowEmpty, "ui_browserShowEmpty", "1", CVAR_ARCHIVE },

	{ &ui_brassTime, "cg_brassTime", "2500", CVAR_ARCHIVE },
	{ &ui_drawCrosshair, "cg_drawCrosshair", "4", CVAR_ARCHIVE },
	{ &ui_drawCrosshairNames, "cg_drawCrosshairNames", "1", CVAR_ARCHIVE },
	{ &ui_marks, "cg_marks", "1", CVAR_ARCHIVE },

	{ &ui_server1, "server1", "", CVAR_ARCHIVE },
	{ &ui_server2, "server2", "", CVAR_ARCHIVE },
	{ &ui_server3, "server3", "", CVAR_ARCHIVE },
	{ &ui_server4, "server4", "", CVAR_ARCHIVE },
	{ &ui_server5, "server5", "", CVAR_ARCHIVE },
	{ &ui_server6, "server6", "", CVAR_ARCHIVE },
	{ &ui_server7, "server7", "", CVAR_ARCHIVE },
	{ &ui_server8, "server8", "", CVAR_ARCHIVE },
	{ &ui_server9, "server9", "", CVAR_ARCHIVE },
	{ &ui_server10, "server10", "", CVAR_ARCHIVE },
	{ &ui_server11, "server11", "", CVAR_ARCHIVE },
	{ &ui_server12, "server12", "", CVAR_ARCHIVE },
	{ &ui_server13, "server13", "", CVAR_ARCHIVE },
	{ &ui_server14, "server14", "", CVAR_ARCHIVE },
	{ &ui_server15, "server15", "", CVAR_ARCHIVE },
	{ &ui_server16, "server16", "", CVAR_ARCHIVE },

	{ &ui_cdkeychecked, "ui_cdkeychecked", "0", CVAR_ROM },

// Hypothermia: Enhanced UI cvars
	{ &uie_gametype, "uie_gametype", "0", CVAR_ARCHIVE | CVAR_ROM },
	{ &uie_pure, "uie_pure", "1", CVAR_ARCHIVE | CVAR_ROM },
	{ &uie_dedicated, "uie_dedicated", "0", CVAR_ARCHIVE | CVAR_ROM },
	{ &uie_inactivity, "uie_inactivity", "90", CVAR_ARCHIVE | CVAR_ROM },
	{ &uie_animsfx, "uie_s_animsfx", "1", CVAR_ARCHIVE },
	{ &uie_mapicons, "uie_mapicons", "0", CVAR_ARCHIVE|CVAR_ROM },
	{ &uie_autoclosebotmenu, "uie_autoclosebotmenu", "0", CVAR_ARCHIVE|CVAR_ROM },
	{ &uie_config_showid, "uie_config_showid", "0", CVAR_ARCHIVE|CVAR_ROM },
	{ &uie_pmove_fixed, "uie_pmove_fixed", "1", CVAR_ARCHIVE|CVAR_ROM },
	{ &uie_pmove_msec, "uie_pmove_msec", "8", CVAR_ARCHIVE|CVAR_ROM },
	{ &uie_smoothclients, "uie_smoothclients", "1", CVAR_ARCHIVE|CVAR_ROM },
	{ &uie_syncclients, "uie_syncclients", "0", CVAR_ARCHIVE|CVAR_ROM },
	{ &uie_allowmaxrate, "uie_allowmaxrate", "0", CVAR_ARCHIVE|CVAR_ROM },
	{ &uie_maxrate, "uie_maxrate", "0", CVAR_ARCHIVE|CVAR_ROM },
	{ &uie_allowdownload, "uie_allowdownload", "0", CVAR_ARCHIVE|CVAR_ROM },
	{ &uie_allowpass, "uie_allowpass", "0", CVAR_ARCHIVE|CVAR_ROM },
	{ &uie_password, "uie_password", "", CVAR_ARCHIVE|CVAR_ROM },
	{ &uie_allowvote, "uie_allowvote", "0", CVAR_ARCHIVE|CVAR_ROM },
	{ &uie_minPing, "uie_minPing", "0", CVAR_ARCHIVE|CVAR_ROM },
	{ &uie_maxPing, "uie_maxPing", "0", CVAR_ARCHIVE|CVAR_ROM },
	{ &uie_allowMinPing, "uie_allowMinPing", "0", CVAR_ARCHIVE|CVAR_ROM },
	{ &uie_allowMaxPing, "uie_allowMaxPing", "0", CVAR_ARCHIVE|CVAR_ROM },
	{ &uie_preventConfigBug, "uie_preventConfigBug", "1", CVAR_ARCHIVE|CVAR_ROM },
	{ &uie_map_multisel, "uie_map_multisel", "0", CVAR_ARCHIVE|CVAR_ROM },
	{ &uie_map_list, "uie_map_list", "0", CVAR_ARCHIVE|CVAR_ROM },
	{ &uie_bot_multisel, "uie_bot_multisel", "0", CVAR_ARCHIVE|CVAR_ROM },
	{ &uie_bot_list, "uie_bot_list", "0", CVAR_ARCHIVE|CVAR_ROM },
	{ &uie_ingame_dynamicmenu, "uie_ingame_dynamicmenu", "1", CVAR_ARCHIVE|CVAR_ROM },

	{ &uie_ffa_fragtype, "uie_ffa_fragtype", "0", CVAR_ARCHIVE },
	{ &uie_ffa_timetype, "uie_ffa_timetype", "0", CVAR_ARCHIVE },
	{ &uie_ffa_customfraglimits, "uie_ffa_customfraglimits", "", CVAR_ARCHIVE | CVAR_ROM },
	{ &uie_ffa_customtimelimits, "uie_ffa_customtimelimits", "", CVAR_ARCHIVE | CVAR_ROM },
	{ &uie_ffa_maplist, "uie_ffa_maplist", "", CVAR_ARCHIVE | CVAR_ROM },
	{ &uie_ffa_maplistexclude, "uie_ffa_maplistexclude", "", CVAR_ARCHIVE | CVAR_ROM },
	{ &uie_ffa_MapRepeat, "uie_ffa_MapRepeat", "1", CVAR_ARCHIVE | CVAR_ROM },
	{ &uie_ffa_MapSource, "uie_ffa_MapSource", "0", CVAR_ARCHIVE | CVAR_ROM },
	{ &uie_ffa_RandomMapCount, "uie_ffa_RandomMapCount", "5", CVAR_ARCHIVE | CVAR_ROM },
	{ &uie_ffa_RandomMapType, "uie_ffa_RandomMapType", "0", CVAR_ARCHIVE | CVAR_ROM },
	{ &uie_ffa_slottype, "uie_ffa_slottype", "", CVAR_ARCHIVE | CVAR_ROM },
	{ &uie_ffa_botname, "uie_ffa_botname", "", CVAR_ARCHIVE | CVAR_ROM },
	{ &uie_ffa_botexclude, "uie_ffa_botexclude", "", CVAR_ARCHIVE | CVAR_ROM },
	{ &uie_ffa_botskill, "uie_ffa_botskill", "", CVAR_ARCHIVE | CVAR_ROM },
	{ &uie_ffa_BotSelection, "uie_ffa_BotSelection", "0", CVAR_ARCHIVE | CVAR_ROM },
	{ &uie_ffa_BotCount, "uie_ffa_BotCount", "4", CVAR_ARCHIVE | CVAR_ROM },
	{ &uie_ffa_BotChange, "uie_ffa_BotChange", "0", CVAR_ARCHIVE | CVAR_ROM },
	{ &uie_ffa_OpenSlotCount, "uie_ffa_OpenSlotCount", "0", CVAR_ARCHIVE | CVAR_ROM },
	{ &uie_ffa_BotSkillType, "uie_ffa_BotSkillType", "0", CVAR_ARCHIVE | CVAR_ROM },
	{ &uie_ffa_BotSkillValue, "uie_ffa_BotSkillValue", "0", CVAR_ARCHIVE | CVAR_ROM },
	{ &uie_ffa_BotSkillBias, "uie_ffa_BotSkillBias", "0", CVAR_ARCHIVE | CVAR_ROM },
	{ &uie_ffa_PlayerJoinAs, "uie_ffa_PlayerJoinAs", "0", CVAR_ARCHIVE | CVAR_ROM },
	{ &uie_ffa_hostname, "uie_ffa_hostname", "UIE Deathmatch server", CVAR_ARCHIVE | CVAR_ROM },
	{ &uie_ffa_ForceRespawn, "uie_ffa_ForceRespawn", "30", CVAR_ARCHIVE | CVAR_ROM },
	{ &uie_ffa_itemGroups, "uie_ffa_itemGroups", "", CVAR_ARCHIVE | CVAR_ROM },
	{ &uie_ffa_itemsHidden, "uie_ffa_itemsHidden", "", CVAR_ARCHIVE | CVAR_ROM },
	{ &uie_ffa_WarmUp, "uie_ffa_WarmUp", "20", CVAR_ARCHIVE | CVAR_ROM },
	{ &uie_ffa_doWarmUp, "uie_ffa_doWarmUp", "0", CVAR_ARCHIVE | CVAR_ROM },

	{ &uie_tourney_fragtype, "uie_tourney_fragtype", "0", CVAR_ARCHIVE | CVAR_ROM },
	{ &uie_tourney_timetype, "uie_tourney_timetype", "0", CVAR_ARCHIVE | CVAR_ROM },
	{ &uie_tourney_customfraglimits, "uie_tourney_customfraglimits", "", CVAR_ARCHIVE | CVAR_ROM },
	{ &uie_tourney_customtimelimits, "uie_tourney_customtimelimits", "", CVAR_ARCHIVE | CVAR_ROM },
	{ &uie_tourney_maplist, "uie_tourney_maplist", "", CVAR_ARCHIVE | CVAR_ROM },
	{ &uie_tourney_maplistexclude, "uie_tourney_maplistexclude", "", CVAR_ARCHIVE | CVAR_ROM },
	{ &uie_tourney_MapRepeat, "uie_tourney_MapRepeat", "1", CVAR_ARCHIVE | CVAR_ROM },
	{ &uie_tourney_MapSource, "uie_tourney_MapSource", "0", CVAR_ARCHIVE | CVAR_ROM },
	{ &uie_tourney_RandomMapCount, "uie_tourney_RandomMapCount", "5", CVAR_ARCHIVE | CVAR_ROM },
	{ &uie_tourney_RandomMapType, "uie_tourney_RandomMapType", "0", CVAR_ARCHIVE | CVAR_ROM },
	{ &uie_tourney_slottype, "uie_tourney_slottype", "", CVAR_ARCHIVE | CVAR_ROM },
	{ &uie_tourney_botname, "uie_tourney_botname", "", CVAR_ARCHIVE | CVAR_ROM },
	{ &uie_tourney_botexclude, "uie_tourney_botexclude", "", CVAR_ARCHIVE | CVAR_ROM },
	{ &uie_tourney_botskill, "uie_tourney_botskill", "", CVAR_ARCHIVE | CVAR_ROM },
	{ &uie_tourney_BotSelection, "uie_tourney_BotSelection", "0", CVAR_ARCHIVE | CVAR_ROM },
	{ &uie_tourney_BotCount, "uie_tourney_BotCount", "4", CVAR_ARCHIVE | CVAR_ROM },
	{ &uie_tourney_BotChange, "uie_tourney_BotChange", "0", CVAR_ARCHIVE | CVAR_ROM },
	{ &uie_tourney_OpenSlotCount, "uie_tourney_OpenSlotCount", "0", CVAR_ARCHIVE | CVAR_ROM },
	{ &uie_tourney_BotSkillType, "uie_tourney_BotSkillType", "0", CVAR_ARCHIVE | CVAR_ROM },
	{ &uie_tourney_BotSkillValue, "uie_tourney_BotSkillValue", "0", CVAR_ARCHIVE | CVAR_ROM },
	{ &uie_tourney_BotSkillBias, "uie_tourney_BotSkillBias", "0", CVAR_ARCHIVE | CVAR_ROM },
	{ &uie_tourney_PlayerJoinAs, "uie_tourney_PlayerJoinAs", "0", CVAR_ARCHIVE | CVAR_ROM },
	{ &uie_tourney_hostname, "uie_tourney_hostname", "UIE Tournament server", CVAR_ARCHIVE | CVAR_ROM },
	{ &uie_tourney_ForceRespawn, "uie_tourney_ForceRespawn", "30", CVAR_ARCHIVE | CVAR_ROM },
	{ &uie_tourney_itemGroups, "uie_tourney_itemGroups", "", CVAR_ARCHIVE | CVAR_ROM },
	{ &uie_tourney_itemsHidden, "uie_tourney_itemsHidden", "", CVAR_ARCHIVE | CVAR_ROM },
	{ &uie_tourney_WarmUp, "uie_tourney_WarmUp", "20", CVAR_ARCHIVE | CVAR_ROM },
	{ &uie_tourney_doWarmUp, "uie_tourney_doWarmUp", "1", CVAR_ARCHIVE | CVAR_ROM },

	{ &uie_team_fragtype, "uie_team_fragtype", "0", CVAR_ARCHIVE | CVAR_ROM },
	{ &uie_team_timetype, "uie_team_timetype", "0", CVAR_ARCHIVE | CVAR_ROM },
	{ &uie_team_customfraglimits, "uie_team_customfraglimits", "", CVAR_ARCHIVE | CVAR_ROM },
	{ &uie_team_customtimelimits, "uie_team_customtimelimits", "", CVAR_ARCHIVE | CVAR_ROM },
	{ &uie_team_maplist, "uie_team_maplist", "", CVAR_ARCHIVE | CVAR_ROM },
	{ &uie_team_maplistexclude, "uie_team_maplistexclude", "", CVAR_ARCHIVE | CVAR_ROM },
	{ &uie_team_MapRepeat, "uie_team_MapRepeat", "1", CVAR_ARCHIVE | CVAR_ROM },
	{ &uie_team_MapSource, "uie_team_MapSource", "0", CVAR_ARCHIVE | CVAR_ROM },
	{ &uie_team_RandomMapCount, "uie_team_RandomMapCount", "5", CVAR_ARCHIVE | CVAR_ROM },
	{ &uie_team_RandomMapType, "uie_team_RandomMapType", "0", CVAR_ARCHIVE | CVAR_ROM },
	{ &uie_team_slottype, "uie_team_slottype", "", CVAR_ARCHIVE | CVAR_ROM },
	{ &uie_team_botname, "uie_team_botname", "", CVAR_ARCHIVE | CVAR_ROM },
	{ &uie_team_botexclude, "uie_team_botexclude", "", CVAR_ARCHIVE | CVAR_ROM },
	{ &uie_team_botexclude, "uie_team_botexclude", "", CVAR_ARCHIVE | CVAR_ROM },
	{ &uie_team_botskill, "uie_team_botskill", "", CVAR_ARCHIVE | CVAR_ROM },
	{ &uie_team_BotSelection, "uie_team_BotSelection", "0", CVAR_ARCHIVE | CVAR_ROM },
	{ &uie_team_BotCount, "uie_team_BotCount", "4", CVAR_ARCHIVE | CVAR_ROM },
	{ &uie_team_BotChange, "uie_team_BotChange", "0", CVAR_ARCHIVE | CVAR_ROM },
	{ &uie_team_OpenSlotCount, "uie_team_OpenSlotCount", "0", CVAR_ARCHIVE | CVAR_ROM },
	{ &uie_team_BotSkillType, "uie_team_BotSkillType", "0", CVAR_ARCHIVE | CVAR_ROM },
	{ &uie_team_BotSkillValue, "uie_team_BotSkillValue", "0", CVAR_ARCHIVE | CVAR_ROM },
	{ &uie_team_BotSkillBias, "uie_team_BotSkillBias", "0", CVAR_ARCHIVE | CVAR_ROM },
	{ &uie_team_PlayerJoinAs, "uie_team_PlayerJoinAs", "0", CVAR_ARCHIVE | CVAR_ROM },
	{ &uie_team_TeamSwapped, "uie_team_TeamSwapped", "0", CVAR_ARCHIVE | CVAR_ROM },
	{ &uie_team_hostname, "uie_team_hostname", "UIE Team Deathmatch server", CVAR_ARCHIVE | CVAR_ROM },
	{ &uie_team_ForceRespawn, "uie_team_ForceRespawn", "30", CVAR_ARCHIVE | CVAR_ROM },
	{ &uie_team_AutoJoin, "uie_team_AutoJoin", "0", CVAR_ARCHIVE | CVAR_ROM },
	{ &uie_team_TeamBalance, "uie_team_TeamBalance", "0", CVAR_ARCHIVE | CVAR_ROM },
	{ &uie_team_itemGroups, "uie_team_itemGroups", "", CVAR_ARCHIVE | CVAR_ROM },
	{ &uie_team_itemsHidden, "uie_team_itemsHidden", "", CVAR_ARCHIVE | CVAR_ROM },
	{ &uie_team_WarmUp, "uie_team_WarmUp", "20", CVAR_ARCHIVE | CVAR_ROM },
	{ &uie_team_doWarmUp, "uie_team_doWarmUp", "0", CVAR_ARCHIVE | CVAR_ROM },

	{ &uie_ctf_capturetype, "uie_ctf_capturetype", "0", CVAR_ARCHIVE | CVAR_ROM },
	{ &uie_ctf_timetype, "uie_ctf_timetype", "0", CVAR_ARCHIVE | CVAR_ROM },
	{ &uie_ctf_customcapturelimits, "uie_ctf_customcapturelimits", "", CVAR_ARCHIVE | CVAR_ROM },
	{ &uie_ctf_customtimelimits, "uie_ctf_customtimelimits", "", CVAR_ARCHIVE | CVAR_ROM },
	{ &uie_ctf_maplist, "uie_ctf_maplist", "", CVAR_ARCHIVE | CVAR_ROM },
	{ &uie_ctf_maplistexclude, "uie_ctf_maplistexclude", "", CVAR_ARCHIVE | CVAR_ROM },
	{ &uie_ctf_MapRepeat, "uie_ctf_MapRepeat", "1", CVAR_ARCHIVE | CVAR_ROM },
	{ &uie_ctf_MapSource, "uie_ctf_MapSource", "0", CVAR_ARCHIVE | CVAR_ROM },
	{ &uie_ctf_RandomMapCount, "uie_ctf_RandomMapCount", "5", CVAR_ARCHIVE | CVAR_ROM },
	{ &uie_ctf_RandomMapType, "uie_ctf_RandomMapType", "0", CVAR_ARCHIVE | CVAR_ROM },
	{ &uie_ctf_slottype, "uie_ctf_slottype", "", CVAR_ARCHIVE | CVAR_ROM },
	{ &uie_ctf_botname, "uie_ctf_botname", "", CVAR_ARCHIVE | CVAR_ROM },
	{ &uie_ctf_botskill, "uie_ctf_botskill", "", CVAR_ARCHIVE | CVAR_ROM },
	{ &uie_ctf_BotSelection, "uie_ctf_BotSelection", "0", CVAR_ARCHIVE | CVAR_ROM },
	{ &uie_ctf_BotCount, "uie_ctf_BotCount", "4", CVAR_ARCHIVE | CVAR_ROM },
	{ &uie_ctf_BotChange, "uie_ctf_BotChange", "0", CVAR_ARCHIVE | CVAR_ROM },
	{ &uie_ctf_OpenSlotCount, "uie_ctf_OpenSlotCount", "0", CVAR_ARCHIVE | CVAR_ROM },
	{ &uie_ctf_BotSkillType, "uie_ctf_BotSkillType", "0", CVAR_ARCHIVE | CVAR_ROM },
	{ &uie_ctf_BotSkillValue, "uie_ctf_BotSkillValue", "0", CVAR_ARCHIVE | CVAR_ROM },
	{ &uie_ctf_BotSkillBias, "uie_ctf_BotSkillBias", "0", CVAR_ARCHIVE | CVAR_ROM },
	{ &uie_ctf_PlayerJoinAs, "uie_ctf_PlayerJoinAs", "0", CVAR_ARCHIVE | CVAR_ROM },
	{ &uie_ctf_TeamSwapped, "uie_ctf_TeamSwapped", "0", CVAR_ARCHIVE | CVAR_ROM },
	{ &uie_ctf_hostname, "uie_ctf_hostname", "UIE CTF server", CVAR_ARCHIVE | CVAR_ROM },
	{ &uie_ctf_ForceRespawn, "uie_ctf_ForceRespawn", "30", CVAR_ARCHIVE | CVAR_ROM },
	{ &uie_ctf_AutoJoin, "uie_ctf_AutoJoin", "0", CVAR_ARCHIVE | CVAR_ROM },
	{ &uie_ctf_TeamBalance, "uie_ctf_TeamBalance", "0", CVAR_ARCHIVE | CVAR_ROM },
	{ &uie_ctf_itemGroups, "uie_ctf_itemGroups", "", CVAR_ARCHIVE | CVAR_ROM },
	{ &uie_ctf_itemsHidden, "uie_ctf_itemsHidden", "", CVAR_ARCHIVE | CVAR_ROM },
	{ &uie_ctf_WarmUp, "uie_ctf_WarmUp", "20", CVAR_ARCHIVE | CVAR_ROM },
	{ &uie_ctf_doWarmUp, "uie_ctf_doWarmUp", "0", CVAR_ARCHIVE | CVAR_ROM }
// HypoThermia: end Enhanced UI cvars
};

int		cvarTableSize = sizeof(cvarTable) / sizeof(cvarTable[0]);


/*
=================
UI_RegisterCvars
=================
*/
void UI_RegisterCvars( void ) {
	int			i;
	cvarTable_t	*cv;

	for ( i = 0, cv = cvarTable ; i < cvarTableSize ; i++, cv++ ) {
		trap_Cvar_Register( cv->vmCvar, cv->cvarName, cv->defaultString, cv->cvarFlags );
	}

	// we also set default values for the disable_* Cvars here,
	// since first usage defines their default value
	UIE_StartServer_RegisterDisableCvars(qtrue);
}

/*
=================
UI_UpdateCvars
=================
*/
void UI_UpdateCvars( void ) {
	int			i;
	cvarTable_t	*cv;

	for ( i = 0, cv = cvarTable ; i < cvarTableSize ; i++, cv++ ) {
		trap_Cvar_Update( cv->vmCvar );
	}
}



/*
=================
UI_IsValidCvar
=================
*/
qboolean UI_IsValidCvar(const char* cvar)
{
	int i;

	for (i = 0; i < cvarTableSize; i++)
	{
		if (!Q_stricmp(cvar, cvarTable[i].cvarName))
			return qtrue;
	}
	return qfalse;
}



